/* -*-c-*-
 *
 * $Id: tripe.c,v 1.14 2004/04/18 18:08:11 mdw Exp $
 *
 * Main program
 *
 * (c) 2001 Straylight/Edgeware
 */

/*----- Licensing notice --------------------------------------------------* 
 *
 * This file is part of Trivial IP Encryption (TrIPE).
 *
 * TrIPE is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * TrIPE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with TrIPE; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*----- Header files ------------------------------------------------------*/

#include "tripe.h"

/*----- Global variables --------------------------------------------------*/

sel_state sel;

/*----- Static variables --------------------------------------------------*/

static sel_timer it;
#define T_INTERVAL MIN(1)

/*----- Main code ---------------------------------------------------------*/

/* --- @interval@ --- *
 *
 * Arguments:	@struct timeval *tv@ = time when called
 *		@void *v@ = boring pointer
 *
 * Returns:	---
 *
 * Use:		Called periodically to do housekeeping tasks.
 */

void interval(struct timeval *tv, void *v)
{
  struct timeval tvv;
  T( trace(T_PEER, "peer: interval timer"); )
  rand_seed(RAND_GLOBAL, MAXHASHSZ);
  p_interval();
  tvv = *tv;
  tvv.tv_sec += T_INTERVAL;
  sel_addtimer(&sel, &it, &tvv, interval, v);
}

/* --- @main@ --- *
 *
 * Arguments:	@int argc@ = number of command line arguments
 *		@char *argv[]@ = vector of arguments
 *
 * Returns:	Zero if OK, nonzero on error.
 *
 * Use:		Main program.  Provides a simple VPN.
 */

static void usage(FILE *fp)
{
  pquis(fp, "Usage: $ [-D] [-d dir] [-b addr] [-p port]\n\
	[-U user] [-G group] [-a socket] [-T trace-opts]\n\
	[-k priv-keyring] [-K pub-keyring] [-t key-tag]\n");
}

static void version(FILE *fp)
{
  pquis(fp, "$, version " VERSION "\n");
}

static void help(FILE *fp)
{
  version(fp);
  fputc('\n', fp);
  usage(fp);
  fputs("\n\
Options:\n\
\n\
-h, --help		Display this help text.\n\
-v, --version		Display version number.\n\
-u, --usage		Display pointless usage message.\n\
\n\
-D, --daemon		Run in the background.\n\
-d, --directory=DIR	Switch to directory DIR [default " CONFIGDIR "].\n\
-b, --bind-address=ADDR	Bind UDP socket to this IP ADDR.\n\
-p, --port=PORT		Select UDP port to listen to.\n\
-U, --setuid=USER	Set uid to USER after initialization.\n\
-G, --setgid=GROUP	Set gid to GROUP after initialization.\n\
-k, --priv-keyring=FILE	Get private key from FILE.\n\
-K, --pub-keyring=FILE	Get public keys from FILE.\n\
-t, --tag=KEYTAG	Use private key labelled TAG.\n\
-a, --admin-socket=FILE	Use FILE as the adminstration socket.\n\
" T( "\
-T, --trace=OPTIONS	Turn on tracing options.\n\
" ) "\
", fp);
}

int main(int argc, char *argv[])
{
  const char *kr_priv = "keyring", *kr_pub = "keyring.pub";
  const char *tag_priv = "tripe-dh";
  const char *csock = SOCKETDIR "/tripesock";
  const char *dir = CONFIGDIR;
  const char *p;
  unsigned port = 0;
  struct in_addr baddr = { INADDR_ANY };
  unsigned f = 0;
  uid_t u = -1;
  gid_t g = -1;

#define f_bogus 1u
#define f_daemon 2u

  ego(argv[0]);
  T( trace_on(stderr, 0); )

  if ((p = getenv("TRIPEDIR")) != 0)
    dir = p;

  for (;;) {
    static const struct option opts[] = {
      { "help",		0,		0,	'h' },
      { "version",	0,		0,	'v' },
      { "usage",	0,		0,	'u' },

      { "daemon",	0,		0,	'D' },
      { "uid",		OPTF_ARGREQ,	0,	'U' },
      { "setuid",	OPTF_ARGREQ,	0,	'U' },
      { "gid",		OPTF_ARGREQ,	0,	'G' },
      { "setgid",	OPTF_ARGREQ,	0,	'G' },
      { "bind-address",	OPTF_ARGREQ,	0,	'b' },
      { "port",		OPTF_ARGREQ,	0,	'p' },
      { "directory",	OPTF_ARGREQ,	0,	'd' },
      { "priv-keyring",	OPTF_ARGREQ,	0,	'k' },
      { "pub-keyring",	OPTF_ARGREQ,	0,	'K' },
      { "tag",		OPTF_ARGREQ,	0,	't' },
      { "admin-socket", OPTF_ARGREQ,	0,	'a' },
#ifndef NTRACE
      { "trace",	OPTF_ARGREQ,	0,	'T' },
#endif

      { "tunnel",	0,		0,	'0' },
      { 0,		0,		0,	0 }
    };

    int i = mdwopt(argc, argv, "hvu DU:G: b:p:d:k:K:t:a:" T("T:"),
		   opts, 0, 0, 0);
    if (i < 0)
      break;
    switch (i) {
      case 'h':
	help(stdout);
	exit(0);
      case 'v':
	version(stdout);
	exit(0);
      case 'u':
	usage(stdout);
	exit(0);

      case 'D':
	f |= f_daemon;
	break;
      case 'U': {
	struct passwd *pw;
	char *p;
	unsigned long i = strtoul(optarg, &p, 0);
	if (!*p)
	  pw = getpwuid(i);
	else
	  pw = getpwnam(optarg);
	if (!pw)
	  die(EXIT_FAILURE, "user `%s' not found", optarg);
	u = pw->pw_uid;
	if (g == -1)
	  g = pw->pw_gid;
      } break;
      case 'G': {
	struct group *gr;
	char *p;
	unsigned long i = strtoul(optarg, &p, 0);
	if (!*p)
	  gr = getgrgid(i);
	else
	  gr = getgrnam(optarg);
	if (!gr)
	  die(EXIT_FAILURE, "group `%s' not found", optarg);
	g = gr->gr_gid;
      } break;

      case 'b': {
	struct hostent *h = gethostbyname(optarg);
	if (!h)
	  die(EXIT_FAILURE, "unknown host name `%s'", optarg);
	memcpy(&baddr, h->h_addr, sizeof(struct in_addr));
      } break;
      case 'p': {
	char *p;
	unsigned long i = strtoul(optarg, &p, 0);
	if (*p) {
	  struct servent *s = getservbyname(optarg, "udp");
	  if (!s)
	    die(EXIT_FAILURE, "unknown service name `%s'", optarg);
	  i = ntohs(s->s_port);
	}
	if (i == 0 || i >= 65536)
	  die(EXIT_FAILURE, "bad port number %lu", i);
	port = i;
      } break;
      case 'd':
	dir = optarg;
	break;
      case 'k':
	kr_priv = optarg;
	break;
      case 'K':
	kr_pub = optarg;
	break;
      case 'a':
	csock = optarg;
	break;
      case 't':
	tag_priv = optarg;
	break;
#ifndef NTRACE
      case 'T':
	tr_flags = traceopt(tr_opts, optarg, tr_flags, 0);
	trace_level(tr_flags);
	break;
#endif
      case '0': {
	static const char *tun[] = { "notdef", "unet", "bsd", "linux" };
	puts(tun[TUN_TYPE]);
	exit(0);
      } break;
      default:
	f |= f_bogus;
	break;
    }
  }

  if (optind < argc || (f & f_bogus)) {
    usage(stderr);
    exit(EXIT_FAILURE);
  }

  if (chdir(dir)) {
    die(EXIT_FAILURE, "can't set current directory to `%s': %s",
	dir, strerror(errno));
  }

  sel_init(&sel);
  sig_init(&sel);
  rand_noisesrc(RAND_GLOBAL, &noise_source);
  rand_seed(RAND_GLOBAL, MAXHASHSZ);
  signal(SIGPIPE, SIG_IGN);
  tun_init();
  p_init(baddr, port);
  if (!(f & f_daemon))
    a_create(STDIN_FILENO, STDOUT_FILENO);
  if (g != (gid_t)-1) {
    if (setgid(g) || (getuid() == 0 && setgroups(1, &g))) {
      die(EXIT_FAILURE, "couldn't setgid to %u: %s",
	  (unsigned)g, strerror(errno));
    }
  }
  if (u != (uid_t)-1) {
    if (setuid(u)) {
      die(EXIT_FAILURE, "couldn't setuid to %u: %s",
	  (unsigned)u, strerror(errno));
    }
  }
  km_init(kr_priv, kr_pub, tag_priv);
  a_init(csock);
  if (f & f_daemon) {
    if (u_daemon())
      die(EXIT_FAILURE, "couldn't become a daemon: %s", strerror(errno));
    a_daemon();
  }

  {
    struct timeval tv;
    tv.tv_sec = time(0) + T_INTERVAL;
    tv.tv_usec = 0;
    sel_addtimer(&sel, &it, &tv, interval, 0);
  }

  {
    int selerr = 0;
    for (;;) {
      if (!sel_select(&sel))
        selerr = 0;
      else if (errno != EINTR && errno != EAGAIN) {
	a_warn("select failed: %s", strerror(errno));
	selerr++;
	if (selerr > 8) {
	  a_warn("too many select errors: bailing out");
	  a_quit();
	}
      }
    }
  }

  return (0);
}

/*----- That's all, folks -------------------------------------------------*/
